use Test2::V0;
use Test2::Plugin::Times;
use Test2::Plugin::ExitSummary;
use Test2::Plugin::NoWarnings;
use FindBin qw($Bin);
use lib "$Bin/../lib";


plan(1);
BEGIN {
    use Clearbuilt::ExcelErator;
    pass('Module loads correctly.');
}